import csv as stdcsv
from io import BytesIO
from typing import Generator

import openpyxl
import xlrd


def generate_dialect(inspection: dict) -> stdcsv.Dialect:
    class CustomDialect(stdcsv.unix_dialect):
        # TODO: it would be nice to have more info from csvdetective to feed the dialect
        # in the meantime we might want to sniff the file a bit
        delimiter = inspection["separator"]

    return CustomDialect()


class Reader:
    def __init__(self, file_path, inspection):
        self.file_path = file_path
        self.inspection = inspection
        self.nb_skip = self.inspection["header_row_idx"]
        self.mapping = {
            "openpyxl": "iter_rows",
            "xlrd": "get_rows",
        }
        self.nb_columns = len(self.inspection["header"])
        self.reader = None

    def __enter__(self):
        if self.inspection.get("engine") == "openpyxl":
            with open(self.file_path, "rb") as f:
                content = BytesIO(f.read())
            self.file = openpyxl.load_workbook(content)
            self.sheet = self.file[self.inspection["sheet_name"]]
            self.reader = self._excel_reader()

        elif self.inspection.get("engine") == "xlrd":
            self.file = xlrd.open_workbook(self.file_path)
            self.sheet = self.file[self.inspection["sheet_name"]]
            self.reader = self._excel_reader()

        else:
            self.file = open(self.file_path, encoding=self.inspection["encoding"])
            self.reader = stdcsv.reader(
                self._skip_rows(), dialect=generate_dialect(self.inspection)
            )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.file is not None and hasattr(self.file, "close"):
            self.file.close()

    def _skip_rows(self):
        # skipping header
        for _ in range(self.nb_skip + 1):
            next(self.file)
        return self.file

    def _excel_reader(self) -> Generator:
        _method = getattr(self.sheet, self.mapping[self.inspection["engine"]])
        for idx, row in enumerate(_method()):
            # skipping header
            if idx <= self.nb_skip:
                continue
            yield [c.value for c in row]

    def __iter__(self):
        return self.reader



import json

from marshmallow import Schema, fields


class CheckSchema(Schema):
    check_id = fields.Integer(data_key="id")
    catalog_id = fields.Integer()
    url = fields.Str()
    domain = fields.Str()
    created_at = fields.DateTime()
    check_status = fields.Integer(data_key="status")
    headers = fields.Function(lambda obj: json.loads(obj["headers"]) if obj["headers"] else {})
    timeout = fields.Boolean()
    response_time = fields.Float()
    error = fields.Str()
    dataset_id = fields.Str()
    resource_id = fields.UUID()
    next_check_at = fields.DateTime()
    deleted = fields.Boolean()
    parsing_started_at = fields.DateTime()
    parsing_finished_at = fields.DateTime()
    parsing_error = fields.Str()
    parsing_table = fields.Str()
    parquet_url = fields.Str()
    parquet_size = fields.Integer()
    pmtiles_url = fields.Str()
    pmtiles_size = fields.Integer()

    def create(self, data):
        return self.load(data)


class CheckGroupBy(Schema):
    value = fields.Str()
    count = fields.Integer()



import json
import uuid

from aiohttp import web
from asyncpg import Record
from asyncpg.exceptions import UniqueViolationError

from udata_hydra import config
from udata_hydra.db.resource import Resource
from udata_hydra.db.resource_exception import ResourceException
from udata_hydra.schemas import ResourceExceptionSchema


async def get_all_resources_exceptions(request: web.Request) -> web.Response:
    """Endpoint to get all resource exceptions from the DB
    Respond with a 200 status code with a list of resource exceptions
    """
    resources_exceptions: list[Record] = await ResourceException.get_all()

    return web.json_response(
        [ResourceExceptionSchema().dump(dict(r)) for r in resources_exceptions]
    )


async def create_resource_exception(request: web.Request) -> web.Response:
    """Endpoint to receive a resource exception creation event from a source
    Will create a new exception in the DB "resources_exceptions" table.
    Respond with a 201 status code with the created resource exception
    """
    try:
        payload = await request.json()
        resource_id: str = payload["resource_id"]
        table_indexes: dict[str, str] | None = payload.get("table_indexes")
        # format should be like this:
        #    {
        #       "column_name": "index_type",
        #       "column_name": "index_type"
        #       ...
        #    },
        comment: str | None = payload.get("comment")
    except Exception as err:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(err)}))

    try:
        resource_exception: Record = await ResourceException.insert(
            resource_id=resource_id,
            table_indexes=table_indexes,
            comment=comment,
        )
    except ValueError as err:
        raise web.HTTPBadRequest(text=f"Resource exception could not be created: {str(err)}")
    except UniqueViolationError:
        raise web.HTTPBadRequest(text="Resource exception already exists")

    return web.json_response(ResourceExceptionSchema().dump(dict(resource_exception)), status=201)


async def update_resource_exception(request: web.Request) -> web.Response:
    """Endpoint to update a resource exception in the DB
    Respond with a 200 status code with the updated resource exception
    """
    try:
        resource_id = str(uuid.UUID(request.match_info["resource_exception_id"]))
    except Exception as e:
        raise web.HTTPBadRequest(text=f"error: {str(e)}")

    resource: Record | None = await Resource.get(resource_id)
    if not resource:
        raise web.HTTPNotFound()

    try:
        payload = await request.json()
        table_indexes: dict[str, str] | None = payload.get("table_indexes")
        comment: str | None = payload.get("comment")
        if table_indexes:
            valid, error = ResourceExceptionSchema.are_table_indexes_valid(table_indexes)
            if not valid:
                raise web.HTTPBadRequest(text=error)
    except Exception as err:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(err)}))

    resource_exception: Record = await ResourceException.update(
        resource_id=resource_id,
        table_indexes=table_indexes,
        comment=comment,
    )

    return web.json_response(ResourceExceptionSchema().dump(dict(resource_exception)))


async def delete_resource_exception(request: web.Request) -> web.Response:
    """Endpoint to delete a resource exception from the DB
    Respond with a 204 status code
    """
    try:
        resource_id = str(uuid.UUID(request.match_info["resource_exception_id"]))
    except Exception as e:
        raise web.HTTPBadRequest(text=f"error: {str(e)}")

    resource: Record | None = await Resource.get(resource_id)
    if not resource:
        raise web.HTTPNotFound()

    await ResourceException.delete(resource_id)

    return web.HTTPNoContent()



import logging
import os
import subprocess
from datetime import datetime, timezone

import tippecanoe
from asyncpg import Record

from udata_hydra import config
from udata_hydra.analysis import helpers
from udata_hydra.db.check import Check
from udata_hydra.db.resource import Resource
from udata_hydra.db.resource_exception import ResourceException
from udata_hydra.utils import (
    IOException,
    ParseException,
    Timer,
    handle_parse_exception,
)
from udata_hydra.utils.minio import MinIOClient

log = logging.getLogger("udata-hydra")
minio_client = MinIOClient(bucket=config.MINIO_PMTILES_BUCKET, folder=config.MINIO_PMTILES_FOLDER)


async def analyse_geojson(
    check: dict,
    file_path: str | None = None,
) -> None:
    """Launch GeoJSON analysis from a check or an URL (debug), using previously downloaded file at file_path if any"""
    if not config.GEOJSON_TO_PMTILES:
        log.debug("GEOJSON_TO_PMTILES turned off, skipping.")
        return

    resource_id: str = str(check["resource_id"])
    url = check["url"]

    # Update resource status to ANALYSING_GEOJSON
    resource: Record | None = await Resource.update(resource_id, {"status": "ANALYSING_GEOJSON"})

    # Check if the resource is in the exceptions table
    exception: Record | None = await ResourceException.get_by_resource_id(resource_id)

    timer = Timer("analyse-geojson")
    assert any(_ is not None for _ in (check["id"], url))

    tmp_file = None
    try:
        tmp_file = await helpers.read_or_download_file(
            check=check,
            file_path=file_path,
            file_format="geojson",
            exception=exception,
        )
        timer.mark("download-file")

        check = await Check.update(check["id"], {"parsing_started_at": datetime.now(timezone.utc)})

        # Convert to PMTiles
        try:
            pmtiles_url, pmtiles_size = await geojson_to_pmtiles(
                file_path=tmp_file.name,
                resource_id=resource_id,
            )
            timer.mark("geojson-to-pmtiles")
        except Exception as e:
            raise ParseException(
                step="pmtiles_export", resource_id=resource_id, url=url, check_id=check["id"]
            ) from e

        check = await Check.update(
            check["id"],
            {
                "parsing_finished_at": datetime.now(timezone.utc),
                "pmtiles_url": pmtiles_url,
                "pmtiles_size": pmtiles_size,
            },
        )

    except (ParseException, IOException) as e:
        await handle_parse_exception(e, None, check)
    finally:
        await helpers.notify_udata(resource, check)
        timer.stop()
        if tmp_file is not None:
            tmp_file.close()
            os.remove(tmp_file.name)

        # Reset resource status to None
        await Resource.update(resource_id, {"status": None})


async def geojson_to_pmtiles(
    file_path: str,
    resource_id: str | None = None,
) -> tuple[str, int]:
    """
    Convert a GeoJSON file to PMTiles format.

    Args:
        file_path: GeoJSON file path to convert.
        resource_id: Optional resource ID for status updates.

    Returns:
        pmtiles_url: URL of the PMTiles file.
        pmtiles_size: size of the PMTiles file.
    """

    log.debug(f"Converting GeoJSON to PMTiles for {file_path}")

    if resource_id:
        await Resource.update(resource_id, {"status": "CONVERTING_TO_PMTILES"})

    output_pmtiles = f"{resource_id}.pmtiles"

    command = [
        "--maximum-zoom=g",  # guess
        "-o",
        output_pmtiles,
        "--coalesce-densest-as-needed",
        "--extend-zooms-if-still-dropping",
        file_path,
    ]
    exit_code = tippecanoe._program("tippecanoe", *command)
    if exit_code:
        raise ValueError(f"GeoJSON to PMTiles conversion failed for {file_path}")
    log.debug(f"Successfully converted {file_path} to {output_pmtiles}")

    pmtiles_size = os.path.getsize(output_pmtiles)
    pmtiles_url: str = minio_client.send_file(output_pmtiles)

    return pmtiles_url, pmtiles_size



import json


def detect_geojson_from_headers(check: dict) -> bool:
    headers: dict = json.loads(check["headers"] or "{}")
    # in some cases geojson files have the content-type `application/json`
    # but adding this in the list would not have been a restrictive enough condition
    # so we check the URL, which is satisfactory for now
    if any(
        headers.get("content-type", "").lower().startswith(ct)
        for ct in ["application/vnd.geo+json"]
    ) or "geojson" in check.get("url", ""):
        return True
    return False
